import math
import numpy as np

distH = 0.1
alphaH = np.arccos(-1/3.0)
s6 = 0.5 * np.sqrt(3.0)

def setup(xAI,xAJ,xAK):
    rij = 0.0
    ra = 0.0
    
    sij = []
    sb = []
    sa = []
    
    for i in range(3):
        xd = xAJ[i]
        sij.append(xAI[i]-xd)
        sb.append(xd-xAK[i])
        rij+=np.square(sij[i])
        #print rij
    rij = np.sqrt(rij)
    #print sij
    
    sa.append(sij[1]*sb[2]-sij[2]*sb[1])
    sa.append(sij[2]*sb[0]-sij[0]*sb[2])
    sa.append(sij[0]*sb[1]-sij[1]*sb[0])
    #print sa
    #import sys
    #sys.exit()
    for i in range(3):
        sij[i]=sij[i]/rij
        ra+=np.square(sa[i])
    
    ra = np.sqrt(ra)
    #print "ra=",ra
    #import sys
    #sys.exit()
    for i in range(3):
        sa[i]=sa[i]/ra
    
    # Important to remember to reset sb here
    sb = []
    sb.append(sa[1]*sij[2]-sa[2]*sij[1])
    sb.append(sa[2]*sij[0]-sa[0]*sij[2])
    sb.append(sa[0]*sij[1]-sa[1]*sij[0])
    
    #print "sb=",sb
    #import sys
    #sys.exit()
    return sa,sb,sij

def one_single_atom(xAI,xAJ,xAK):
    
    xH1 = []
    #xAI=[-0.0677, -0.123, -0.0491]
    #xAJ=[-0.0001,0.0064,-0.0491]
    #xAK=[0.1499,-0.011,-0.0491]
    #print xAI
    #print xAJ
    #print xAK
    sa,sb,sij = setup(xAI,xAJ,xAK)
    
    for i in range(3):
        xH1.append(xAI[i]+distH*np.sin(alphaH)*sb[i]-distH*np.cos(alphaH)*sij[i])
    
    return xH1

def two_atoms(xAI,xAJ,xAK):
    r1 = []
    r2 = []
    r2 = []
    
    xH1 = []
    xH2 = []
    
    for i in range(3):
        r1.append(xAI[i]-0.5*(xAJ[i]+xAK[i]))
        
    b = np.linalg.norm(r1)
    r2 = np.subtract(xAI,xAJ)
    r3 = np.subtract(xAI,xAK)
    r4 = np.cross(r2,r3)
    n  = np.linalg.norm(r4)
    
    for i in range(3):
        xH1.append(xAI[i]+distH*(np.cos(alphaH/2.0*r1[i]/b+np.sin(alphaH/2.0)*r4[i]/n)))
        xH2.append(xAI[i]+distH*(np.cos(alphaH/2.0*r1[i]/b-np.sin(alphaH/2.0)*r4[i]/n)))
        
    return xH1, xH2

def three_atoms(xAI,xAJ,xAK):
    xH1 = []
    xH2 = []
    xH3 = []
    
    sa,sb,sij = setup(xAI,xAJ,xAK)
    
    for i in range(3):
        xH1.append(xAI[i]+distH*np.sin(alphaH)*sb[i]-distH*np.cos(alphaH)*sij[i])
        
        xH2.append(xAI[i]-
                   distH*np.sin(alphaH)*0.5*sb[i] +
                   distH*np.sin(alphaH)*s6*sa[i] -
                   distH*np.cos(alphaH)*sij[i])
        
        xH3.append(xAI[i]-
                   distH*np.sin(alphaH)*0.5*sb[i] -
                   distH*np.sin(alphaH)*s6*sa[i] -
                   distH*np.cos(alphaH)*sij[i])
        
    return xH1,xH2,xH3

# Compare two lists of atoms    
def check_atoms(target_atoms,atoms):
    s = list(set(target_atoms)-set(atoms))
    s1 = list(set(atoms)-set(target_atoms))
    
    return s,s1


# Return the residue name for atom ai in list l                                
def get_resn(l,ai):
    resn = ""
    
    for i in range(len(l)):
        try:
            if l[i] and l[i][0].strip()==ai:
                resn = l[i][1]
        except:
            print "Error in get_resn(): ai=%s",ai
            sys.exit()
            
    return resn
        
def get_resi(l,ai):
    resi = -1
    
    for i in range(len(l)):
        try:
            if l[i] and l[i][0].strip()==ai:
                resi = l[i][2]
        except:
            print "Error in get_resi(): ai=%s",ai
            sys.exit()

    return resi

def get_chain(l,ai):
    chain = ""
    
    for i in range(len(l)):
        try:
            if l[i] and l[i][0].strip()==ai:
                chain = l[i][3]
        except:
            print "Error in get_chain(): ai=%s"%ai
            sys.exit()
            
    return chain


# Return the position in list for atom ai in list l                            
def get_pos_in_list(l,ai):
    
    pos = -1
    
    for i in range(len(l)):
        try:
            if l[i] and l[i][0].strip()==ai:
                pos = i
        except TypeError:
            print "Error in get_pos_in_list(): ai=%s"%ai
            sys.exit()

    return pos

# Return the xyz coordinates of the atom in ai in list l                       
def get_xyz_coords(l,ai):
    
    xyz = np.zeros((3,))
    
    for i in range(len(l)):
        try:
            if l[i] and l[i][0].strip()==ai:

                xyz[0]=l[i][3][0]
                xyz[1]=l[i][3][1]
                xyz[2]=l[i][3][2]
        except TypeError:
            print "Error in get_xyz_coords(): ai=%s"%ai
            sys.exit()
            
    return xyz



#/usr/bin/env python
import sys
import os
from structure import Protein
from transform import transform
from convert import convert

def read_input(input,longresnum,debug=0):
    # Read in the input structure - pdb or gro based on file ending
    struct = Protein(input,longresnum,debug=0)
    return struct

def transf(struct,ffin,ffout):
    t = transform()
    t.read_transforms()
    new_struct = t.do(struct,ffin,ffout)
    return new_struct

def conv(struct,ffin,lin,lout,n):
    c = convert()
    c.read_conversions()
    new_struct = c.do(struct,ffin,lin,lout,n)
    return new_struct




#!/usr/bin/env python
import argparse
import lipid_conv.lipid_conv as lipid_conv

# Use argparse to parse command line options                                 
parser = argparse.ArgumentParser(description='Lipid-converter')
parser.add_argument('-f','--input',required=True)
parser.add_argument('-o','--output',required=True)
parser.add_argument('-mode','--mode',default='transform',help='Transform or Convert')
parser.add_argument('-ffin','--ffin',default='berger',help='Source force field')
parser.add_argument('-ffout','--ffout',default='charmm36',help='Target force field')
parser.add_argument('-lin','--lin',help='Input lipid')
parser.add_argument('-lout','--lout',help='Output lipid')
parser.add_argument('-n','--n',help='Convert every n-th lipid')
parser.add_argument('-canonical','--canonical',action='store_true',help='Canonical force field sorting of output')
parser.add_argument('-longresnum','--longresnum',action='store_true',help='Assume long residues numbers in pdb file')

# Get the arguments
args = parser.parse_args()
print args

# Read in the input structure - pdb or gro based on file ending 
struct = lipid_conv.read_input(input=args.input,
                               longresnum=args.longresnum,
                               debug=0)

# Do conversion or transformations
if args.mode == 'transform':
    new_struct = lipid_conv.transf(struct=struct,
                                   ffin=args.ffin,
                                   ffout=args.ffout)
elif args.mode == 'convert':
    new_struct = lipid_conv.conv(struct=struct,
                                 ffin=args.ffin,
                                 lin=args.lin,
                                 lout=args.lout,
                                 n = args.n)
else:
    print "Either transform or convert please...!"
    sys.exit()
        
if args.canonical:
    # If we are changing from one forcefield to another, we are              
    # sorting according to the output ff                                     
    if args.mode == 'transform':
        ff_sort = args.ffout
        
    # And similarily, if we are changing a lipid type within a forcefield,   
    # we sort on this ff (ie the input ff)                                   
    if args.mode == 'convert':
        ff_sort = args.ffin
        
    # We also need a special trick for amber/lipids11, since the residue     
    # names are identical between different lipids (the plug-and-play        
    # architecture differentiates between different lipid headgroups and     
    # tails, so that different lipids with the same tails will have the same
    # residue names.                                                         
    # We therefore loop over the input structure and save an array with the  
    # old residue names                                                      
    # Todo: work out something when going from an amber/lipid11 coordinate   
    # file (probably some kind of third file with residue names)             
    resmap = None
    if ff_sort == 'lipid11':
        resmap = struct.get_resnames()
        
    new_struct = new_struct.sort(ff_sort,resmap)

# Write out result                                                           
new_struct.write(args.output)



import math
import re
import os
import sys
import numpy as np
from forcefields import ff_sortings,lipid11_residues

Vsites = ['MN1','MN2']
directive = re.compile('^ *\[ *(.*) *\]')

class Sort:
    def __init__(self):
        self.sorts = {}

    def sort(self,ff,resmap):
        # Get everything
        self.read_sortings()
        
        new = Protein()
        
        for resnum in self.get_residues():
            residue = self.get_residue_data(resnum)

            # amber/lipid11 
            if resmap is not None:
                # Need to subtract 1 from resnum since we start
                # with 0-indexing in python, but most likely not
                # in the pdb-file
                resname = resmap[resnum-1]
            else:
                resname = residue[0][1]

            out = self.sort_residue(residue,ff,resname)
            #print "APA"
            #print out
            #print ff
            new.add_residue_data(out)

        return new
        
    def sort_residue(self,residue,ff,resname=None):

        # Get the name of the residue to be sorted
        if resname:
            resn = resname
        else:
            #resname = residue[0][1]
            resn = residue[0][1]
        #print residue[0]
        #print ff,resn
        sort_data = self.sorts[ff,resn]['atoms']
        
        out = [0]*len(sort_data)
        #print sort_data
        if ff == 'lipid11':

            for i,(ai_atname,ai_resname) in enumerate(sort_data):
                
                found = 0
                for j in range(len(residue)):
                    aj_atname = residue[j][0]
                    aj_resname = residue[j][1]
                    
                    if ( (ai_atname == aj_atname) and
                         (ai_resname == aj_resname) ):
                                
                        found = 1
                        out[i]=(aj_atname,
                                aj_resname,
                                residue[j][2],
                                residue[j][3])
                        
                        #sorted[j]=1
                        # Some lipids in lipid11 have two identical tails
                        # (eg. DPPC) we cannot simply compare two lists
                        # with each other. We therefore delete it from the 
                        # list so that it cannot be "found" again, and also
                        # then need to break out of the loop
                        residue.pop(j)
                        break
                
                if found == 0:
                    print "Could not sort atom %s in residue %s"%(ai_atname,resn)
        else:
            for i,ai in enumerate(sort_data):
                ai = sort_data[i][0]
            
                for j in range(len(residue)):
                    aj = residue[j][0]
                
                    if ai == aj:
                    #print ai,aj
                        out[i]=(aj,
                                residue[j][1],
                                residue[j][2],
                                residue[j][3])
        #print out
        
        #sys.exit()
        return out
                
    def read_sortings(self):
        for ff in ff_sortings:
            fn = ""
            
            try:
                path = os.path.dirname(__file__)
                path = os.path.join(path,ff,'sortings.top')
                fn = open(path,'r')
            except:
                print "Could not open sortings file for %s"%ff

            atoms = []
            lipid = ""
            
            for line in fn:
                s = line.strip()
                
                if s.startswith("["):
                    d = re.findall(directive,s)[0].strip().lower()

                    # If we get to end of an entry, store it 
                    if d=='end':

                        # Create a new dict for this entry 
                        m = {}
                        m['atoms']=atoms
                        #print ff,lipid
                        self.sorts[ff,lipid]=m
                        # Reset atoms  
                        atoms = []
                        
                    continue
                if not s:
                    continue
                
                elif d == 'atoms':
                    atoms.append(s.split())

                elif d == 'molecule':
                    lipid = s.split()[0]


class Protein(Sort):
    def __init__(self,file_in=None,longresnum=False,debug=0):
        self.title=''
        self.atcounter = 0
        
        self.header = []
        self.footer = []
        
        self.label = []
        self.atnum = []
        self.elem = []
        self.mass = []
        self.atname = []
        self.atalt = []
        self.resname = []
        self.chain = []
        self.resnum = []
        self.resext = []
        self.coord = []
        self.occ = []
        self.b = []
        self.sequence = {}
        self.box = []
        self.velocity = []
        
        self.debug = debug
        
        # Read in the file based on its extension
        if file_in:
            filetype = os.path.splitext(file_in)[1]
            
            gro = re.compile('.gro')
            pdb = re.compile('.pdb')
            
            if gro.match(filetype):
                self.read_gro(file_in)
            elif pdb.match(filetype):
                self.read_pdb(file_in,longresnum)
            else:
                print "Unknown file-type"
                sys.exit()
                
        # We also need to inititate the Sort base class here
        # This is a bit weird, but I'm a noob at oop anyway :-)
        Sort.__init__(self)

    def read_pdb(self,file_in,longresnum,debug=0):
        lines = file(file_in).readlines()
        self.read_pdb_lines(lines,longresnum,debug)
        
    def read_pdb_lines(self,lines,longresnum,debug):
        atom_hetatm = re.compile('(ATOM  |HETATM)')
      
        for line in lines:
            if atom_hetatm.match(line):
                
                line = line[:-1]
                # We throw away a bunch of stuff in the input
                # pdb file here... this could be changed of course
                if line[12:16].strip() not in Vsites:
                    self.atname.append(line[12:16].strip())
                    self.resname.append(line[17:21].strip())
                    if longresnum:
                        self.resnum.append(int(line[22:27]))
                    else:
                        self.resnum.append(int(line[22:26]))
                    self.coord.append((float(line[30:38])/10, float(line[38:46])/10, float(line[46:54])/10))
                    
                    self.atcounter += 1
                
        if debug:
            return len(self.atnum),self.atcounter
                            
    def read_gro(self,file_in,debug=0):
        lines = file(file_in).readlines()
        self.read_gro_lines(lines,debug)
        
    def read_gro_lines(self,lines,debug):
        self.title = lines[0][:-1]
        # We first set atcounter to whatever it is in the input file
        self.atcounter = int(lines[1][:-1])

        if self.debug:
            print self.title
            print self.atcounter
            print lines[2][:-1],len(lines[2])
 
        ctr = 0
        # and use that value to read in all atoms
        for line in lines[2:self.atcounter+3][:-1]:
            if line[10:15].strip() not in Vsites:
                self.resnum.append(int(line[0:5]))
                self.resname.append(line[5:10].strip())
                self.atname.append(line[10:15].strip())
                self.atnum.append(int(line[15:20]))
                
                first_decimal = line.index('.')
                second_decimal = line[first_decimal+1:].index('.')
                #print first_decimal
                #print second_decimal
                incr = second_decimal + 1
                #print incr
                #print line[20:20+incr]
                #print line[28:20+2*incr]
                self.coord.append((float(line[20:20+incr]), float(line[28:20+2*incr]), float(line[36:20+3*incr])))
                
                ctr = ctr + 1

        self.atcounter = ctr

    def get_residues(self):
        res = np.unique(self.resnum)
        return res

    def get_resnames(self):
        resnames = []
        for resnum in self.get_residues():
            resnames.append(self.resname[resnum])
            
        return resnames

    # For lipid11, we need a routine that makes sense of the PA/PC/OL
    # nomenclature, and translates into a single residue name
    # translation table in lipid11_residues
    def get_canonical_residue_name(self,residue):
        resnames = []
        
        for i  in range(len(residue)):
            resn = residue[i][1]
            resnames.append(resn)

        # By converting to a set we can compare two lists
        # as long as we don't care about order (which we don't here)
        unique_resnames = set(resnames)
        #unique_resnames = list(unique_resnames)
        
        resn_out = ""

        for r in lipid11_residues:
            l11_res = set(lipid11_residues[r])
            
            if l11_res == unique_resnames:
                resn_out = r
         
        #print len(resn_out)
        #print resn_out
        #sys.exit()
        if len(resn_out)>0:
            return resn_out
        else:
            print "Could not find a canonical residue name for the"
            print "[%s] combination of residue names in lipid11"%', '.join(map(str, list(unique_resnames)))
            sys.exit()
    
    def get_residue_data(self,resi):
        out = []
        
        for i in range(self.atcounter):
            if self.resnum[i]==resi:
                
                out.append((#self.atnum[i],
                            self.atname[i],
                            #self.atalt[i],
                            self.resname[i],
                            #self.chain[i],
                            self.resnum[i],
                            #self.resext[i],
                            self.coord[i]))
                            #self.occ[i],
                            #self.b[i]))
                
        return out
                
    def add_residue_data(self,residue):
        
        for i in range(len(residue)):
            #print residue[i]
            #if residue[i] == 0
            atname = residue[i][0]
            resname = residue[i][1]
            resnum = residue[i][2]
            coords = residue[i][3]

            self.resnum.append(resnum)
            self.resname.append(resname)
            self.atname.append(atname)
            self.coord.append(coords)

            self.atcounter = self.atcounter + 1
            
    def write_pdb(self,file_out):
        f = open(file_out,'w')

        for i in range(self.atcounter):
            atnum = i % 99999
            self.write_pdb_line(f,i,atnum)

        f.close()

    def write_pdb_line(self,file_out,i,atnum):
        label = 'ATOM  '
        atalt = ' '
        chain = ' '
        resext = ' '
        occ = 1.0
        b = 0.0
        elem = ''
        blank = ''

        # This isn't the strict pdb format string, but it lets us
        # write pdb-files that have 4-letter long reside codes
        # and 5-numbered residue numbers
        file_out.write('%-6s%5d %4s%1s%-4s %4d    %8.3f%8.3f%8.3f%6.2f%6.2f          %2s%2s\n'%(label,atnum+1,self.atname[i],atalt,self.resname[i],self.resnum[i],self.coord[i][0]*10.,self.coord[i][1]*10.,self.coord[i][2]*10.,occ,b,blank,elem)) 
        
    def write_gro(self,file_out):
        f = open(file_out,'w')
        
        f.write('lipid-converter gro-file\n')
        f.write('%d\n'%self.atcounter)

        for i in range(self.atcounter):
            atnum = i % 99999
            self.write_gro_line(f,i,atnum)

        f.write('%10.5f%10.5f%10.5f'%(0,0,0))
        f.close()

    def write_gro_line(self,file_out,i,atnum):
        file_out.write("%5d%-5s%5s%5d%8.3f%8.3f%8.3f\n"%(self.resnum[i]%1e5,self.resname[i],self.atname[i],atnum+1,self.coord[i][0],self.coord[i][1],self.coord[i][2]))

    def write(self,filename):
        filetype = os.path.splitext(filename)[1]
        
        gro = re.compile('.gro')
        pdb = re.compile('.pdb')
        
        if gro.match(filetype):
            self.write_gro(filename)
        elif pdb.match(filetype):
            self.write_pdb(filename)
        else:
            print "Unknown filetype in write"
            sys.exit()





import os
import glob
import re
import sys
from structure import Protein
from forcefields import ff_conversions
import aux

regexp = re.compile('^ *\[ *(.*) *\]')

class convert():
    def __init__(self):
        self.conversions = {}
        
    def read_conversions(self):
        for ff in ff_conversions:
            fn = ""

            try:
                path = os.path.dirname(__file__)
                path = os.path.join(path,ff,'conversions.top')
                fn = open(path,'r')
            except:
                print "Could not open convert file for %s"%ff
                
            add_atoms = []
            remove = []
            rename = []
            
            lin = ""
            lout = ""

            for line in fn:
                l = line.strip()

                if l.startswith("["):
                    d = re.findall(regexp,l)[0].strip().lower()
                    
                    # If we get to end of an entry, store it
                    if d=='end':
                        
                        # Create a new dict for this entry
                        m = {}
                        m['add']=add_atoms
                        m['remove']=remove
                        m['rename']=rename
                        
                        # Add this combination of ff,lin and lout
                        # as a tuple key
                        self.conversions[ff,lin,lout]=m

                        # Reset these three
                        rename = []
                        add_atoms = []
                        remove = []

                    continue

                if not l:
                    continue

                elif d == 'rename':
                    rename.append(l.split())
                    
                elif d == 'add':
                    add_atoms.append(l.split())
                    
                elif d == 'remove':
                    remove.append(l.split())

                elif d == 'molecule':
                    lin = l.split()[0]
                    
                elif d =='target':
                    lout = l.split()[0]

    def do(self,prot,ff,lin,lout,n):

        new = Protein()

        # Get the total number of residues
        total_resnum = len(prot.get_residues())
        
        convert_count = 0
        for resnum in prot.get_residues():
            residue = prot.get_residue_data(resnum)
            resname = residue[0][1]
            
            if resname == lin:
                convert_count = convert_count + 1

            # Is this a residue we want to change?
            if convert_count%n==0 and resname == lin:
                try:
                    lipid = residue[0][1]
                    add_atoms = self.conversions[ff,lin,lout]['add']
                    rename_atoms = self.conversions[ff,lin,lout]['rename']
                    remove_atoms = self.conversions[ff,lin,lout]['remove']
                except:
                    print "No conversion between %s and %s in %s found"%(lin,lout,ff)
                    sys.exit()
                    
                # Do the conversions
                transformed = self.remove_atoms(residue,remove_atoms)
                transformed = self.rename_atoms(transformed,rename_atoms)
                transformed = self.build_atoms(transformed,add_atoms)
                transformed = self.update_resname(transformed,lout)
                print "Converted residue %s %d to %s (total numres: %d)"%(lin,resnum,lout,total_resnum)
            else:
                transformed = residue
                
            # Add the result to a new protein   
            #print transformed
            new.add_residue_data(transformed)

        return new
            
    def remove_atoms(self,residue,remove_atoms):
        
        # Make a local copy of the atom names her
        # so we can make operations on the residue 
        # list in the loop
        atoms = [i[0].strip() for i in residue]

        for i in range(len(remove_atoms)):
            ai = remove_atoms[i][0]
            
            for j in range(len(atoms)):
                aj = atoms[j]
                if ai == aj:
                    
                    # This is important since atoms and residue
                    # will be out of sync 
                    pos = aux.get_pos_in_list(residue,aj)
                    residue.pop(pos)
                    
        return residue


    def rename_atoms(self,residue,rename_atoms):

        atoms = [i[0].strip() for i in residue]
               
        for i in range(len(rename_atoms)):
            ai = rename_atoms[i][0]
            aout = rename_atoms[i][1]

            for j in range(len(atoms)):
                aj = atoms[j]

                if ai == aj:
                    #print ai,aout
                    residue[j]=(aout,
                                residue[j][1],
                                residue[j][2],
                                residue[j][3])
        return residue
                    
    def build_atoms(self,residue,add_atoms):
        
        for i in range(len(add_atoms)):
            
            # Get the data for building atoms
            name,suffix,ai,aj,ak = add_atoms[i]
            
            # Number of atoms to build is based on the length of the
            # suffix string
            num = len(suffix)
            
            xai = aux.get_xyz_coords(residue,ai)
            xaj = aux.get_xyz_coords(residue,aj)
            xak = aux.get_xyz_coords(residue,ak)

            pos   = aux.get_pos_in_list(residue,ai)
            resn  = aux.get_resn(residue,ai)
            resi  = aux.get_resi(residue,ai)

            if num==1:
                x1 = aux.one_single_atom(xai,xaj,xak)
                x1_name = name
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))

            elif num==2:
                x1,x2 = aux.two_atoms(xai,xaj,xak)
                x1_name = name+suffix[0]
                x2_name = name+suffix[1]
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))
                residue.insert(pos+2,(x2_name,resn,resi,(x2[0],x2[1],x2[2])))

            elif num==3:
                x1,x2,x3 = aux.three_atoms(xai,xaj,xak)
                x1_name = name+suffix[0]
                x2_name = name+suffix[1]
                x3_name = name+suffix[2]
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))
                residue.insert(pos+2,(x2_name,resn,resi,(x2[0],x2[1],x2[2])))
                residue.insert(pos+2,(x3_name,resn,resi,(x3[0],x3[1],x3[2])))
            else:
                print "Need to specify either 1,2 or 3 hydrogens to construct around central atom %s"%ai
                print "Currently it is %d"%num
                print "Bailing out..."
                
                sys.exit()
                
        return residue
        

    def update_resname(self,residue,lout):
        residue = [list(i) for i in residue]
        
        for i in range(len(residue)):
            residue[i][1]=lout
    
        return residue



import os
import glob
import re
from structure import Protein
from forcefields import ff_transformations
import aux

import sys

regexp = re.compile('^ *\[ *(.*) *\]')

class transform():
    def __init__(self):
        self.transforms = {}
        
    def read_transforms(self):
        for ff in ff_transformations:
            
            fn = ""

            try:
                path = os.path.dirname(__file__)
                path = os.path.join(path,ff,'transforms.top')
                fn = open(path,'r')
            except:
                print "Could not open transform file for %s"%ff
            
            transf = []
            hyd = []
            ffout = ""
            lipid = ""

            for line in fn:
                l = line.strip()
                
                if l.startswith("["):
                    d = re.findall(regexp,l)[0].strip().lower()
                    
                    # If we get to end of an entry, store it
                    if d=='end':
                        
                        # Create a new dict for this entry
                        m = {}
                        m['transf']=transf
                        if hyd:
                            m['hyd']=hyd
                        
                            # Add this combination of lipid, ff and ffout
                            # as a tuple key
                        self.transforms[lipid,ff,ffout]=m
                        
                        # Reset these two
                        transf = []
                        hyd = []

                    continue

                if not l:
                    continue

                elif d == 'atoms':
                    transf.append(l.split())
                    
                elif d == 'hydrogens':
                    hyd.append(l.split())

                elif d == 'molecule':
                    lipid = l.split()[0]

                elif d =='target':
                    ffout = l.split()[0]

    def do(self,prot,ffin,ffout):

        new = Protein()
        #print self.transforms['POPC','lipid11','charmm36']
        #sys.exit()
        
        for resnum in prot.get_residues():
            residue = prot.get_residue_data(resnum)
            
            print "converting residue %d"%resnum
            try:
                if ffin == 'lipid11':
                    lipid = prot.get_canonical_residue_name(residue)
                else:
                    lipid = residue[0][1]
                    
                transf_atoms = self.transforms[lipid,ffin,ffout]['transf']
                try:
                    hyd = self.transforms[lipid,ffin,ffout]['hyd']
                except KeyError:
                    hyd = ""
            except:
                print "No transformation from %s to %s for residue %s %d found"%(ffin,ffout,lipid,resnum)
                sys.exit()
                
            # Do the transformation
            transformed = self.transform_residue(residue,transf_atoms,ffin,ffout,lipid)
            #print transformed
            #new.add_residue_data(transformed)
            #new.write('test.pdb')
            #sys.exit()
            #print hyd
            #sys.exit()
            if hyd:
                transformed = self.build_atoms(transformed,hyd)
            
            #print transformed
            #sys.exit()
            # Add the result to a new protein
            new.add_residue_data(transformed)
            #new.write('test.pdb')
            #print new.atname
            #print new.coord
            #sys.exit()
        return new
            
    def transform_residue(self,residue,transf_atoms,ffin,ffout,lipid):
        out = []
        #print transf_atoms
        for i in range(len(residue)):
            #print residue
            #sys.exit()
            ain = residue[i][0]
            resn_in = residue[i][1]
            
            #print resn_in
            #sys.exit()
            
            # Amber/lipids11 is a special case in that we also
            # need to change (and later sort on) the residue name
            # as well                                              
            # note that in this case the transf_atom structure       
            # from read_tranforms() will contain the new residue     
            # name as well in the [2]-column                         
            # this pertains to when the ffout==lipid11            
            if ffin == 'lipid11':
                for j in range(len(transf_atoms)):
                    if ( (ain == transf_atoms[j][0]) and
                         (resn_in == transf_atoms[j][2]) ):
                        aout = transf_atoms[j][1]
                        resn = lipid
                        
                        resi = residue[i][2]
                        coords = residue[i][3]
                        out.append((aout,resn,resi,coords))
                        break
            elif ffout=='lipid11':
                for j in range(len(transf_atoms)):
                    
                    if(ain == transf_atoms[j][0]):
                        aout = transf_atoms[j][1]
                        resn = transf_atoms[j][2]
                        
                        resi = residue[i][2]
                        coords = residue[i][3]
                        out.append((aout,resn,resi,coords))
                        break
            else:
                for j in range(len(transf_atoms)):
                    
                    if(ain == transf_atoms[j][0]):
                        aout = transf_atoms[j][1]
                        resn = lipid
                        
                        resi = residue[i][2]
                        coords = residue[i][3]
                        out.append((aout,resn,resi,coords))
                        break
        return out

    def build_atoms(self,residue,hyd):
        
        for i in range(len(hyd)):
            
            # Get the data for building atoms
            name,suffix,ai,aj,ak = hyd[i]
            
            # Number of atoms to build is based on the length of the
            # suffix string
            num = len(suffix)
            
            xai = aux.get_xyz_coords(residue,ai)
            xaj = aux.get_xyz_coords(residue,aj)
            xak = aux.get_xyz_coords(residue,ak)

            pos   = aux.get_pos_in_list(residue,ai)
            resn  = aux.get_resn(residue,ai)
            resi  = aux.get_resi(residue,ai)

            if num==1:
                x1 = aux.one_single_atom(xai,xaj,xak)
                x1_name = name
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))

            elif num==2:
                x1,x2 = aux.two_atoms(xai,xaj,xak)
                x1_name = name+suffix[0]
                x2_name = name+suffix[1]
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))
                residue.insert(pos+2,(x2_name,resn,resi,(x2[0],x2[1],x2[2])))

            elif num==3:
                x1,x2,x3 = aux.three_atoms(xai,xaj,xak)
                x1_name = name+suffix[0]
                x2_name = name+suffix[1]
                x3_name = name+suffix[2]
                residue.insert(pos+1,(x1_name,resn,resi,(x1[0],x1[1],x1[2])))
                residue.insert(pos+2,(x2_name,resn,resi,(x2[0],x2[1],x2[2])))
                residue.insert(pos+2,(x3_name,resn,resi,(x3[0],x3[1],x3[2])))
            else:
                print "Need to specify either 1,2 or 3 hydrogens to construct around central atom %s"%ai
                print "Currently it is %d"%num
                print "Bailing out..."
                
                sys.exit()
        #print residue
        return residue
        

    
    



from setuptools import setup

setup(
    name='LipidConverter',
    version='0.1.2',
    author='Per Larsson',
    author_email='larsson.r.per@gmail.com',
    packages=['lipid_conv'],
    package_data={'lipid_conv': ['berger/*.top',
                                 'charmm36/*.top',
                                 'gromos43a1-S3/*.top',
                                 'gromos53a6/*.top',
                                 'gromos54a7/*.top',
                                 'lipid11/*.top',
                                 'opls/*.top']
                  },
    scripts=['bin/lipid_converter.py'],
    url='http://lipid-converter.appspot.com',
    license='LICENSE.txt',
    description='Useful descripton text',
    long_description=open('README.txt').read(),
    install_requires=['NumPy >= 1.6'],
)




# This needs to be updated when a new forcefield is introduced
ff_transformations = ['charmm36',
                      'berger',
                      'gromos43a1-S3',
                      'gromos53a6',
                      'gromos54a7',
                      'opls',
                      'lipid11']

ff_conversions = ['charmm36',
                  'gromos43a1-S3',
                  'berger']

ff_sortings = ['charmm36',
               'berger',
               'gromos43a1-S3',
               'gromos54a7',
               'gromos53a6',
               'opls',
               'lipid11']

# These are all lipids that are defined in the current 
# release of amber/lipid11, and how they are made
# up of different headgroups and tails, as well
# as the corresponding "normal" name in charmm36
lipid11_residues = {'DPPC':['PA','PC','PA'],
                    'DPPE': ['PA','PE','PA'],
                    'DPPS': ['PA','PS','PA'],
                    'DPPG': ['PA','PGR','PA'],
                    'DPPA': ['PA','PH-','PA'],
                    'DOPC': ['OL','PC','OL'],
                    'DOPE': ['OL','PE','OL'],
                    'DOPS': ['OL','PS','OL'],
                    'DOPG': ['OL','PGR','OL'],
                    'DOPA': ['OL','PH-','OL'],
                    'POPC': ['PA','PC','OL'],
                    'POPE': ['PA','PE','OL'],
                    'POPS': ['PA','PS','OL'],
                    'POPG': ['PA','PGR','OL'],
                    'POPA': ['PA','PH-','OL']}
